@@ -1,20 +1,27 @@
-from __future__ import print_function, division, absolute_import
+from __future__ import absolute_import, division, print_function
 
-import aiohttp
 import asyncio
+import io
 import logging
 import re
-import requests
 import weakref
+from copy import copy
 from urllib.parse import urlparse
+
+import aiohttp
+import requests
+
+from fsspec.asyn import AsyncFileSystem, sync, sync_wrapper
+from fsspec.callbacks import _DEFAULT_CALLBACK
+from fsspec.exceptions import FSTimeoutError
 from fsspec.spec import AbstractBufferedFile
-from fsspec.utils import tokenize, DEFAULT_BLOCK_SIZE
-from fsspec.asyn import sync_wrapper, sync, AsyncFileSystem
+from fsspec.utils import DEFAULT_BLOCK_SIZE, nullcontext, tokenize
+
 from ..caching import AllBytes
 
 # https://stackoverflow.com/a/15926317/3821154
-ex = re.compile(r"""<a\s+(?:[^>]*?\s+)?href=(["'])(.*?)\1""")
-ex2 = re.compile(r"""(http[s]?://[-a-zA-Z0-9@:%_+.~#?&/=]+)""")
+ex = re.compile(r"""<(a|A)\s+(?:[^>]*?\s+)?(href|HREF)=["'](?P<url>[^"']+)""")
+ex2 = re.compile(r"""(?P<url>http[s]?://[-a-zA-Z0-9@:%_+.~#?&/=]+)""")
 logger = logging.getLogger("fsspec.http")
 
 
@@ -45,7 +52,8 @@ class HTTPFileSystem(AsyncFileSystem):
         asynchronous=False,
         loop=None,
         client_kwargs=None,
-        **storage_options
+        get_client=get_client,
+        **storage_options,
     ):
         """
         NB: if this is called async, you must await set_client
@@ -66,6 +74,10 @@ class HTTPFileSystem(AsyncFileSystem):
             Passed to aiohttp.ClientSession, see
             https://docs.aiohttp.org/en/stable/client_reference.html
             For example, ``{'auth': aiohttp.BasicAuth('user', 'pass')}``
+        get_client: Callable[..., aiohttp.ClientSession]
+            A callable which takes keyword arguments and constructs
+            an aiohttp.ClientSession. It's state will be managed by
+            the HTTPFileSystem class.
         storage_options: key-value
             Any other parameters passed on to requests
         cache_type, cache_options: defaults used in open
@@ -77,39 +89,70 @@ class HTTPFileSystem(AsyncFileSystem):
         self.cache_type = cache_type
         self.cache_options = cache_options
         self.client_kwargs = client_kwargs or {}
+        self.get_client = get_client
         self.kwargs = storage_options
-        if not asynchronous:
-            self._session = sync(self.loop, get_client, **self.client_kwargs)
-            weakref.finalize(self, sync, self.loop, self.session.close)
-        else:
-            self._session = None
+        self._session = None
 
-    @property
-    def session(self):
-        if self._session is None:
-            raise RuntimeError("please await ``.set_session`` before anything else")
-        return self._session
+        # Clean caching-related parameters from `storage_options`
+        # before propagating them as `request_options` through `self.kwargs`.
+        # TODO: Maybe rename `self.kwargs` to `self.request_options` to make
+        #       it clearer.
+        request_options = copy(storage_options)
+        self.use_listings_cache = request_options.pop("use_listings_cache", False)
+        request_options.pop("listings_expiry_time", None)
+        request_options.pop("max_paths", None)
+        request_options.pop("skip_instance_cache", None)
+        self.kwargs = request_options
+
+        if not asynchronous:
+            sync(self.loop, self.set_session)
+
+    @staticmethod
+    def close_session(loop, session):
+        if loop is not None and loop.is_running():
+            try:
+                sync(loop, session.close, timeout=0.1)
+                return
+            except (TimeoutError, FSTimeoutError):
+                pass
+        connector = getattr(session, "_connector", None)
+        if connector is not None:
+            # close after loop is dead
+            connector._close()
 
     async def set_session(self):
-        self._session = await get_client(**self.client_kwargs)
+        if self._session is None:
+            self._session = await self.get_client(loop=self.loop, **self.client_kwargs)
+            if not self.asynchronous:
+                weakref.finalize(self, self.close_session, self.loop, self._session)
+        return self._session
 
     @classmethod
     def _strip_protocol(cls, path):
         """For HTTP, we always want to keep the full URL"""
         return path
 
-    async def _ls(self, url, detail=True, **kwargs):
+    @classmethod
+    def _parent(cls, path):
+        # override, since _strip_protocol is different for URLs
+        par = super()._parent(path)
+        if len(par) > 7:  # "http://..."
+            return par
+        return ""
+
+    async def _ls_real(self, url, detail=True, **kwargs):
         # ignoring URL-encoded arguments
         kw = self.kwargs.copy()
         kw.update(kwargs)
         logger.debug(url)
-        async with self.session.get(url, **self.kwargs) as r:
-            r.raise_for_status()
+        session = await self.set_session()
+        async with session.get(url, **self.kwargs) as r:
+            self._raise_not_found_for_status(r, url)
             text = await r.text()
         if self.simple_links:
-            links = ex2.findall(text) + ex.findall(text)
+            links = ex2.findall(text) + [u[2] for u in ex.findall(text)]
         else:
-            links = ex.findall(text)
+            links = [u[2] for u in ex.findall(text)]
         out = set()
         parts = urlparse(url)
         for l in links:
@@ -131,7 +174,7 @@ class HTTPFileSystem(AsyncFileSystem):
                     # Ignore FTP-like "parent"
                     out.add("/".join([url.rstrip("/"), l.lstrip("/")]))
         if not out and url.endswith("/"):
-            return await self._ls(url.rstrip("/"), detail=True)
+            out = await self._ls_real(url.rstrip("/"), detail=False)
         if detail:
             return [
                 {
@@ -143,41 +186,124 @@ class HTTPFileSystem(AsyncFileSystem):
             ]
         else:
             return list(sorted(out))
+        return out
 
-    async def _cat_file(self, url, **kwargs):
+    async def _ls(self, url, detail=True, **kwargs):
+
+        if self.use_listings_cache and url in self.dircache:
+            out = self.dircache[url]
+        else:
+            out = await self._ls_real(url, detail=detail, **kwargs)
+            self.dircache[url] = out
+        return out
+
+    ls = sync_wrapper(_ls)
+
+    def _raise_not_found_for_status(self, response, url):
+        """
+        Raises FileNotFoundError for 404s, otherwise uses raise_for_status.
+        """
+        if response.status == 404:
+            raise FileNotFoundError(url)
+        response.raise_for_status()
+
+    async def _cat_file(self, url, start=None, end=None, **kwargs):
         kw = self.kwargs.copy()
         kw.update(kwargs)
         logger.debug(url)
-        async with self.session.get(url, **kw) as r:
-            if r.status == 404:
-                raise FileNotFoundError(url)
-            r.raise_for_status()
+
+        # TODO: extract into testable utility function?
+        if start is not None or end is not None:
+            if start == end:
+                return b""
+            headers = kw.pop("headers", {}).copy()
+
+            headers["Range"] = await self._process_limits(url, start, end)
+            kw["headers"] = headers
+        session = await self.set_session()
+        async with session.get(url, **kw) as r:
             out = await r.read()
+            self._raise_not_found_for_status(r, url)
         return out
 
-    async def _get_file(self, rpath, lpath, chunk_size=5 * 2 ** 20, **kwargs):
+    async def _get_file(
+        self, rpath, lpath, chunk_size=5 * 2 ** 20, callback=_DEFAULT_CALLBACK, **kwargs
+    ):
         kw = self.kwargs.copy()
         kw.update(kwargs)
         logger.debug(rpath)
-        async with self.session.get(rpath, **self.kwargs) as r:
-            if r.status == 404:
-                raise FileNotFoundError(rpath)
-            r.raise_for_status()
+        session = await self.set_session()
+        async with session.get(rpath, **self.kwargs) as r:
+            try:
+                size = int(r.headers["content-length"])
+            except (ValueError, KeyError):
+                size = None
+
+            callback.set_size(size)
+            self._raise_not_found_for_status(r, rpath)
             with open(lpath, "wb") as fd:
                 chunk = True
                 while chunk:
                     chunk = await r.content.read(chunk_size)
                     fd.write(chunk)
+                    callback.relative_update(len(chunk))
+
+    async def _put_file(
+        self,
+        lpath,
+        rpath,
+        chunk_size=5 * 2 ** 20,
+        callback=_DEFAULT_CALLBACK,
+        method="post",
+        **kwargs,
+    ):
+        async def gen_chunks():
+            # Support passing arbitrary file-like objects
+            # and use them instead of streams.
+            if isinstance(lpath, io.IOBase):
+                context = nullcontext(lpath)
+                use_seek = False  # might not support seeking
+            else:
+                context = open(lpath, "rb")
+                use_seek = True
+
+            with context as f:
+                if use_seek:
+                    callback.set_size(f.seek(0, 2))
+                    f.seek(0)
+                else:
+                    callback.set_size(getattr(f, "size", None))
+
+                chunk = f.read(chunk_size)
+                while chunk:
+                    yield chunk
+                    callback.relative_update(len(chunk))
+                    chunk = f.read(chunk_size)
+
+        kw = self.kwargs.copy()
+        kw.update(kwargs)
+        session = await self.set_session()
+
+        method = method.lower()
+        if method not in ("post", "put"):
+            raise ValueError(
+                f"method has to be either 'post' or 'put', not: {method!r}"
+            )
+
+        meth = getattr(session, method)
+        async with meth(rpath, data=gen_chunks(), **kw) as resp:
+            self._raise_not_found_for_status(resp, rpath)
 
     async def _exists(self, path, **kwargs):
         kw = self.kwargs.copy()
         kw.update(kwargs)
         try:
             logger.debug(path)
-            r = await self.session.get(path, **kw)
+            session = await self.set_session()
+            r = await session.get(path, **kw)
             async with r:
                 return r.status < 400
-        except (requests.HTTPError, aiohttp.client_exceptions.ClientError):
+        except (requests.HTTPError, aiohttp.ClientError):
             return False
 
     async def _isfile(self, path, **kwargs):
@@ -191,7 +317,8 @@ class HTTPFileSystem(AsyncFileSystem):
         autocommit=None,  # XXX: This differs from the base class.
         cache_type=None,
         cache_options=None,
-        **kwargs
+        size=None,
+        **kwargs,
     ):
         """Make a file-like object
 
@@ -213,23 +340,24 @@ class HTTPFileSystem(AsyncFileSystem):
         kw = self.kwargs.copy()
         kw["asynchronous"] = self.asynchronous
         kw.update(kwargs)
-        size = self.size(path)
+        size = size or self.info(path, **kwargs)["size"]
+        session = sync(self.loop, self.set_session)
         if block_size and size:
             return HTTPFile(
                 self,
                 path,
-                session=self.session,
+                session=session,
                 block_size=block_size,
                 mode=mode,
                 size=size,
                 cache_type=cache_type or self.cache_type,
                 cache_options=cache_options or self.cache_options,
                 loop=self.loop,
-                **kw
+                **kw,
             )
         else:
             return HTTPStreamFile(
-                self, path, mode=mode, loop=self.loop, session=self.session, **kw
+                self, path, mode=mode, loop=self.loop, session=session, **kw
             )
 
     def ukey(self, url):
@@ -246,25 +374,116 @@ class HTTPFileSystem(AsyncFileSystem):
         which case size will be given as None (and certain operations on the
         corresponding file will not work).
         """
-        size = False
+        info = {}
+        session = await self.set_session()
+
         for policy in ["head", "get"]:
             try:
-                size = await _file_size(
-                    url, size_policy=policy, session=self.session, **self.kwargs
+                info.update(
+                    await _file_info(
+                        url,
+                        size_policy=policy,
+                        session=session,
+                        **self.kwargs,
+                        **kwargs,
+                    )
                 )
-                if size:
+                if info.get("size") is not None:
                     break
-            except Exception:
-                pass
-        else:
-            # get failed, so conclude URL does not exist
-            if size is False:
-                raise FileNotFoundError(url)
-        return {"name": url, "size": size or None, "type": "file"}
+            except Exception as exc:
+                if policy == "get":
+                    # If get failed, then raise a FileNotFoundError
+                    raise FileNotFoundError(url) from exc
+                logger.debug(str(exc))
 
-    def isdir(self, path):
+        return {"name": url, "size": None, **info, "type": "file"}
+
+    async def _glob(self, path, **kwargs):
+        """
+        Find files by glob-matching.
+
+        This implementation is idntical to the one in AbstractFileSystem,
+        but "?" is not considered as a character for globbing, because it is
+        so common in URLs, often identifying the "query" part.
+        """
+        import re
+
+        ends = path.endswith("/")
+        path = self._strip_protocol(path)
+        indstar = path.find("*") if path.find("*") >= 0 else len(path)
+        indbrace = path.find("[") if path.find("[") >= 0 else len(path)
+
+        ind = min(indstar, indbrace)
+
+        detail = kwargs.pop("detail", False)
+
+        if not has_magic(path):
+            root = path
+            depth = 1
+            if ends:
+                path += "/*"
+            elif await self._exists(path):
+                if not detail:
+                    return [path]
+                else:
+                    return {path: await self._info(path)}
+            else:
+                if not detail:
+                    return []  # glob of non-existent returns empty
+                else:
+                    return {}
+        elif "/" in path[:ind]:
+            ind2 = path[:ind].rindex("/")
+            root = path[: ind2 + 1]
+            depth = None if "**" in path else path[ind2 + 1 :].count("/") + 1
+        else:
+            root = ""
+            depth = None if "**" in path else path[ind + 1 :].count("/") + 1
+
+        allpaths = await self._find(
+            root, maxdepth=depth, withdirs=True, detail=True, **kwargs
+        )
+        # Escape characters special to python regex, leaving our supported
+        # special characters in place.
+        # See https://www.gnu.org/software/bash/manual/html_node/Pattern-Matching.html
+        # for shell globbing details.
+        pattern = (
+            "^"
+            + (
+                path.replace("\\", r"\\")
+                .replace(".", r"\.")
+                .replace("+", r"\+")
+                .replace("//", "/")
+                .replace("(", r"\(")
+                .replace(")", r"\)")
+                .replace("|", r"\|")
+                .replace("^", r"\^")
+                .replace("$", r"\$")
+                .replace("{", r"\{")
+                .replace("}", r"\}")
+                .rstrip("/")
+            )
+            + "$"
+        )
+        pattern = re.sub("[*]{2}", "=PLACEHOLDER=", pattern)
+        pattern = re.sub("[*]", "[^/]*", pattern)
+        pattern = re.compile(pattern.replace("=PLACEHOLDER=", ".*"))
+        out = {
+            p: allpaths[p]
+            for p in sorted(allpaths)
+            if pattern.match(p.replace("//", "/").rstrip("/"))
+        }
+        if detail:
+            return out
+        else:
+            return list(out)
+
+    async def _isdir(self, path):
         # override, since all URLs are (also) files
-        return bool(self.ls(path))
+        try:
+            return bool(await self._ls(path))
+        except (FileNotFoundError, ValueError):
+            return False
 
 
 class HTTPFile(AbstractBufferedFile):
@@ -304,7 +523,7 @@ class HTTPFile(AbstractBufferedFile):
         size=None,
         loop=None,
         asynchronous=False,
-        **kwargs
+        **kwargs,
     ):
         if mode != "rb":
             raise NotImplementedError("File mode not supported")
@@ -319,7 +538,7 @@ class HTTPFile(AbstractBufferedFile):
             block_size=block_size,
             cache_type=cache_type,
             cache_options=cache_options,
-            **kwargs
+            **kwargs,
         )
         self.loop = loop
 
@@ -334,11 +553,9 @@ class HTTPFile(AbstractBufferedFile):
             read only part of the data will raise a ValueError.
         """
         if (
-            (length < 0 and self.loc == 0)
-            or (length > (self.size or length))  # explicit read all
-            or (  # read more than there is
-                self.size and self.size < self.blocksize
-            )  # all fits in one block anyway
+            (length < 0 and self.loc == 0)  # explicit read all
+            # but not when the size is known and fits into a block anyways
+            and not (self.size is not None and self.size <= self.blocksize)
         ):
             self._fetch_all()
         if self.size is None:
@@ -354,12 +571,15 @@ class HTTPFile(AbstractBufferedFile):
         This is only called when position is still at zero,
         and read() is called without a byte-count.
         """
+        logger.debug(f"Fetch all for {self}")
         if not isinstance(self.cache, AllBytes):
             r = await self.session.get(self.url, **self.kwargs)
             async with r:
                 r.raise_for_status()
                 out = await r.read()
-                self.cache = AllBytes(out)
+                self.cache = AllBytes(
+                    size=len(out), fetcher=None, blocksize=None, data=out
+                )
                 self.size = len(out)
 
     _fetch_all = sync_wrapper(async_fetch_all)
@@ -372,6 +592,7 @@ class HTTPFile(AbstractBufferedFile):
         and then stream the output - if the data size is bigger than we
         requested, an exception is raised.
         """
+        logger.debug(f"Fetch range for {self}: {start}-{end}")
         kwargs = self.kwargs.copy()
         headers = kwargs.pop("headers", {}).copy()
         headers["Range"] = "bytes=%i-%i" % (start, end - 1)
@@ -385,42 +606,52 @@ class HTTPFile(AbstractBufferedFile):
             if r.status == 206:
                 # partial content, as expected
                 out = await r.read()
-            elif "Content-Length" in r.headers:
-                cl = int(r.headers["Content-Length"])
-                if cl <= end - start:
-                    # data size OK
-                    out = await r.read()
-                else:
-                    raise ValueError(
-                        "Got more bytes (%i) than requested (%i)" % (cl, end - start)
-                    )
+            elif int(r.headers.get("Content-Length", 0)) <= end - start:
+                out = await r.read()
             else:
                 cl = 0
                 out = []
                 while True:
                     chunk = await r.content.read(2 ** 20)
-                    # data size unknown, let's see if it goes too big
+                    # data size unknown, let's read until we have enough
                     if chunk:
                         out.append(chunk)
                         cl += len(chunk)
                         if cl > end - start:
-                            raise ValueError(
-                                "Got more bytes so far (>%i) than requested (%i)"
-                                % (cl, end - start)
-                            )
+                            break
                     else:
                         break
-                out = b"".join(out)
+                out = b"".join(out)[: end - start]
             return out
 
     _fetch_range = sync_wrapper(async_fetch_range)
 
-    def close(self):
-        pass
+    def __reduce__(self):
+        return (
+            reopen,
+            (
+                self.fs,
+                self.url,
+                self.mode,
+                self.blocksize,
+                self.cache.name,
+                self.size,
+            ),
+        )
 
 
-async def get(session, url, **kwargs):
-    return await session.get(url, **kwargs)
+def reopen(fs, url, mode, blocksize, cache_type, size=None):
+    return fs.open(
+        url, mode=mode, block_size=blocksize, cache_type=cache_type, size=size
+    )
+
+
+magic_check = re.compile("([*[])")
+
+
+def has_magic(s):
+    match = magic_check.search(s)
+    return match is not None
 
 
 class HTTPStreamFile(AbstractBufferedFile):
@@ -433,10 +664,16 @@ class HTTPStreamFile(AbstractBufferedFile):
             raise ValueError
         self.details = {"name": url, "size": None}
         super().__init__(fs=fs, path=url, mode=mode, cache_type="none", **kwargs)
-        self.r = sync(self.loop, get, self.session, url, **kwargs)
+
+        async def cor():
+            r = await self.session.get(url, **kwargs).__aenter__()
+            self.fs._raise_not_found_for_status(r, url)
+            return r
+
+        self.r = sync(self.loop, cor)
 
     def seek(self, *args, **kwargs):
-        raise ValueError("Cannot seek strteaming HTTP file")
+        raise ValueError("Cannot seek streaming HTTP file")
 
     async def _read(self, num=-1):
         out = await self.r.content.read(num)
@@ -450,6 +687,10 @@ class HTTPStreamFile(AbstractBufferedFile):
 
     def close(self):
         asyncio.run_coroutine_threadsafe(self._close(), self.loop)
+        super().close()
+
+    def __reduce__(self):
+        return reopen, (self.fs, self.url, self.mode, self.blocksize, self.cache.name)
 
 
 async def get_range(session, url, start, end, file=None, **kwargs):
@@ -469,17 +710,20 @@ async def get_range(session, url, start, end, file=None, **kwargs):
         return out
 
 
-async def _file_size(url, session=None, size_policy="head", **kwargs):
-    """Call HEAD on the server to get file size
+async def _file_info(url, session, size_policy="head", **kwargs):
+    """Call HEAD on the server to get details about the file (size/checksum etc.)
 
     Default operation is to explicitly allow redirects and use encoding
     'identity' (no compression) to get the true size of the target.
     """
+    logger.debug("Retrieve file size for %s" % url)
     kwargs = kwargs.copy()
     ar = kwargs.pop("allow_redirects", True)
     head = kwargs.get("headers", {}).copy()
     head["Accept-Encoding"] = "identity"
-    session = session or await get_client()
+    kwargs["headers"] = head
+
+    info = {}
     if size_policy == "head":
         r = await session.head(url, allow_redirects=ar, **kwargs)
     elif size_policy == "get":
@@ -487,10 +731,29 @@ async def _file_size(url, session=None, size_policy="head", **kwargs):
     else:
         raise TypeError('size_policy must be "head" or "get", got %s' "" % size_policy)
     async with r:
+        r.raise_for_status()
+
+        # TODO:
+        #  recognise lack of 'Accept-Ranges',
+        #                 or 'Accept-Ranges': 'none' (not 'bytes')
+        #  to mean streaming only, no random access => return None
         if "Content-Length" in r.headers:
-            return int(r.headers["Content-Length"])
+            info["size"] = int(r.headers["Content-Length"])
         elif "Content-Range" in r.headers:
-            return int(r.headers["Content-Range"].split("/")[1])
+            info["size"] = int(r.headers["Content-Range"].split("/")[1])
+
+        for checksum_field in ["ETag", "Content-MD5", "Digest"]:
+            if r.headers.get(checksum_field):
+                info[checksum_field] = r.headers[checksum_field]
+
+    return info
+
+
+async def _file_size(url, session=None, *args, **kwargs):
+    if session is None:
+        session = await get_client()
+    info = await _file_info(url, session=session, *args, **kwargs)
+    return info.get("size")
 
 
 file_size = sync_wrapper(_file_size)
