@@ -1,12 +1,15 @@
 import os
-import shutil
 import pickle
-import pytest
+import shutil
 import tempfile
 
+import pytest
+
 import fsspec
-from fsspec.implementations.cached import CachingFileSystem
 from fsspec.compression import compr
+from fsspec.exceptions import BlocksizeMismatchError
+from fsspec.implementations.cached import CachingFileSystem
+
 from .test_ftp import FTPFileSystem
 
 
@@ -178,6 +181,7 @@ def test_pop():
     with pytest.raises(PermissionError):
         fs.pop_from_cache(f1)
     fs.pop_from_cache(f2)
+    fs.pop_from_cache(os.path.join(origin, "uncached-file"))
     assert len(os.listdir(cache2)) == 1
     assert not fs._check_file(f2)
     assert fs._check_file(f1)
@@ -210,10 +214,81 @@ def test_blocksize(ftp_writable):
 
     with fs.open("/out_block", block_size=20) as f:
         assert f.read(1) == b"t"
-    with pytest.raises(ValueError):
+    with pytest.raises(BlocksizeMismatchError):
         fs.open("/out_block", block_size=30)
 
 
+def test_blockcache_multiinstance(ftp_writable):
+    host, port, user, pw = ftp_writable
+    fs = FTPFileSystem(host, port, user, pw)
+    with fs.open("/one", "wb") as f:
+        f.write(b"test" * 40)
+    with fs.open("/two", "wb") as f:
+        f.write(b"test" * 40)
+    fs = fsspec.filesystem(
+        "blockcache",
+        target_protocol="ftp",
+        target_options={"host": host, "port": port, "username": user, "password": pw},
+    )
+
+    with fs.open("/one", block_size=20) as f:
+        assert f.read(1) == b"t"
+    fs2 = fsspec.filesystem(
+        "blockcache",
+        target_protocol="ftp",
+        target_options={"host": host, "port": port, "username": user, "password": pw},
+        skip_instance_cache=True,
+        cache_storage=fs.storage,
+    )
+    assert fs2.cached_files  # loaded from metadata for "one"
+    with fs2.open("/two", block_size=20) as f:
+        assert f.read(1) == b"t"
+    assert "/two" in fs2.cached_files[-1]
+    fs.save_cache()
+    assert list(fs.cached_files[-1]) == ["/one", "/two"]
+    assert list(fs2.cached_files[-1]) == ["/one", "/two"]
+
+
+def test_metadata_save_blocked(ftp_writable, caplog):
+    import logging
+
+    host, port, user, pw = ftp_writable
+    fs = FTPFileSystem(host, port, user, pw)
+    with fs.open("/one", "wb") as f:
+        f.write(b"test" * 40)
+    fs = fsspec.filesystem(
+        "blockcache",
+        target_protocol="ftp",
+        target_options={"host": host, "port": port, "username": user, "password": pw},
+    )
+
+    with fs.open("/one", block_size=20) as f:
+        assert f.read(1) == b"t"
+    fn = os.path.join(fs.storage[-1], "cache")
+    with caplog.at_level(logging.DEBUG):
+        with fs.open("/one", block_size=20) as f:
+            f.seek(21)
+            assert f.read(1)
+            os.remove(fn)
+            os.mkdir(fn)
+    assert "Cache saving failed while closing file" in caplog.text
+    os.rmdir(fn)
+
+    def open_raise(*_, **__):
+        raise NameError
+
+    try:
+
+        with caplog.at_level(logging.DEBUG):
+            with fs.open("/one", block_size=20) as f:
+                fsspec.implementations.cached.open = open_raise
+                f.seek(21)
+                assert f.read(1)
+    finally:
+        fsspec.implementations.cached.__dict__.pop("open", None)
+    assert "Cache save failed due to interpreter shutdown" in caplog.text
+
+
 @pytest.mark.parametrize("impl", ["filecache", "simplecache", "blockcache"])
 def test_local_filecache_creates_dir_if_needed(impl):
     import tempfile
@@ -239,6 +314,35 @@ def test_local_filecache_creates_dir_if_needed(impl):
     assert data_in_cache == data
 
 
+@pytest.mark.parametrize("toplevel", [True, False])
+@pytest.mark.parametrize("impl", ["filecache", "simplecache", "blockcache"])
+def test_get_mapper(impl, toplevel):
+    import tempfile
+
+    original_location = tempfile.mkdtemp()
+    cache_location = tempfile.mkdtemp()
+    os.rmdir(cache_location)
+    original_file = os.path.join(original_location, "afile")
+    data = b"test data"
+    with open(original_file, "wb") as f:
+        f.write(data)
+
+    if toplevel:
+        m = fsspec.get_mapper(
+            f"{impl}::file://{original_location}",
+            **{impl: {"cache_storage": cache_location}},
+        )
+    else:
+        fs = fsspec.filesystem(
+            impl, target_protocol="file", cache_storage=cache_location
+        )
+        m = fs.get_mapper(original_location)
+
+    assert m["afile"] == data
+    assert os.listdir(cache_location)
+    assert m["afile"] == data
+
+
 def test_local_filecache_basic(local_filecache):
     data, original_file, cache_location, fs = local_filecache
 
@@ -343,7 +447,7 @@ def test_filecache_multicache():
     assert fs.cat(f1) == data
 
     assert len(os.listdir(cache1)) == 2  # cache and hashed afile
-    assert len(os.listdir(cache2)) == 0  # hasn't been intialized yet
+    assert len(os.listdir(cache2)) == 0  # hasn't been initialized yet
 
     # populates last cache if file not found in first cache
     fs = fsspec.filesystem(
@@ -426,6 +530,8 @@ def test_filecache_with_checks():
 
     assert fs.cat(f1) == data  # does not change
     assert fs2.cat(f1) == data * 2  # changed, since origin changed
+    with fs2.open(f1) as f:
+        assert f.read() == data * 2  # read also sees new data
     time.sleep(0.11)  # allow cache details to expire
     assert fs.cat(f1) == data * 2  # changed, since origin changed
 
@@ -463,6 +569,25 @@ def test_add_file_to_cache_after_save(local_filecache):
     assert len(fs2.cached_files[-1]) == 1
 
 
+def test_cached_open_close_read(ftp_writable):
+    # Regression test for <https://github.com/fsspec/filesystem_spec/issues/799>
+    host, port, user, pw = ftp_writable
+    fs = FTPFileSystem(host, port, user, pw)
+    with fs.open("/out_block", "wb") as f:
+        f.write(b"test" * 4000)
+    fs = fsspec.filesystem(
+        "cached",
+        target_protocol="ftp",
+        target_options={"host": host, "port": port, "username": user, "password": pw},
+    )
+    with fs.open("/out_block", block_size=1024) as f:
+        pass
+    with fs.open("/out_block", block_size=1024) as f:
+        assert f.read(1) == b"t"
+    # Regression test for <https://github.com/fsspec/filesystem_spec/issues/845>
+    assert fs.cached_files[-1]["/out_block"]["blocks"] == {0}
+
+
 @pytest.mark.parametrize("impl", ["filecache", "simplecache"])
 @pytest.mark.parametrize("compression", ["gzip", "bz2"])
 def test_with_compression(impl, compression):
@@ -478,7 +603,7 @@ def test_with_compression(impl, compression):
         "%s::%s" % (impl, fn),
         "rb",
         compression=compression,
-        **{impl: dict(same_names=True, cache_storage=cachedir)}
+        **{impl: dict(same_names=True, cache_storage=cachedir)},
     ) as f:
         # stores original compressed file, uncompress on read
         assert f.read() == data
@@ -490,7 +615,9 @@ def test_with_compression(impl, compression):
     with fsspec.open(
         "%s::%s" % (impl, fn),
         "rb",
-        **{impl: dict(same_names=True, compression=compression, cache_storage=cachedir)}
+        **{
+            impl: dict(same_names=True, compression=compression, cache_storage=cachedir)
+        },
     ) as f:
         # stores uncompressed data
         assert f.read() == data
@@ -528,7 +655,7 @@ def test_multi_cache(protocol):
     lurl = fsspec.open_local(
         f"{protocol}::memory://file*",
         mode="rb",
-        **{protocol: {"cache_storage": d2, "same_names": True}}
+        **{protocol: {"cache_storage": d2, "same_names": True}},
     )
     assert all(d2 in u for u in lurl)
     assert all(os.path.basename(f) in ["file0", "file1"] for f in lurl)
@@ -538,7 +665,7 @@ def test_multi_cache(protocol):
     lurl = fsspec.open_files(
         f"{protocol}::memory://file*",
         mode="rb",
-        **{protocol: {"cache_storage": d2, "same_names": True}}
+        **{protocol: {"cache_storage": d2, "same_names": True}},
     )
     with lurl as files:
         for f in files:
@@ -589,9 +716,9 @@ def test_multi_cache_chain(protocol):
 def test_strip(protocol):
     fs = fsspec.filesystem(protocol, target_protocol="memory")
     url1 = "memory://afile"
-    assert fs._strip_protocol(url1) == "afile"
-    assert fs._strip_protocol(protocol + "://afile") == "afile"
-    assert fs._strip_protocol(protocol + "::memory://afile") == "afile"
+    assert fs._strip_protocol(url1) == "/afile"
+    assert fs._strip_protocol(protocol + "://afile") == "/afile"
+    assert fs._strip_protocol(protocol + "::memory://afile") == "/afile"
 
 
 @pytest.mark.parametrize("protocol", ["simplecache", "filecache"])
@@ -609,7 +736,7 @@ def test_expiry():
 
     d = tempfile.mkdtemp()
     fs = fsspec.filesystem("memory")
-    fn = "afile"
+    fn = "/afile"
     fn0 = "memory://afile"
     data = b"hello"
     with fs.open(fn0, "wb") as f:
@@ -646,3 +773,51 @@ def test_expiry():
     assert fs.open(fn0, mode="rb").read() == data
     detail, _ = fs._check_file(fn0)
     assert detail["time"] - start_time > 0.09
+
+
+def test_equality():
+    """Test sane behaviour for equality and hashing.
+
+    Make sure that different CachingFileSystem only test equal to each other
+    when they should, and do not test equal to the filesystem they rely upon.
+    Similarly, make sure their hashes differ when they should and are equal
+    when they should not.
+
+    Related: GitHub#577, GitHub#578
+    """
+    from fsspec.implementations.local import LocalFileSystem
+
+    lfs = LocalFileSystem()
+    cfs1 = CachingFileSystem(fs=lfs, cache_storage="raspberry")
+    cfs2 = CachingFileSystem(fs=lfs, cache_storage="banana")
+    cfs3 = CachingFileSystem(fs=lfs, cache_storage="banana")
+    assert cfs1 == cfs1
+    assert cfs1 != cfs2
+    assert cfs1 != cfs3
+    assert cfs2 == cfs3
+    assert cfs1 != lfs
+    assert cfs2 != lfs
+    assert cfs3 != lfs
+    assert hash(lfs) != hash(cfs1)
+    assert hash(lfs) != hash(cfs2)
+    assert hash(lfs) != hash(cfs3)
+    assert hash(cfs1) != hash(cfs2)
+    assert hash(cfs1) != hash(cfs2)
+    assert hash(cfs2) == hash(cfs3)
+
+
+@pytest.mark.xfail
+def test_json():
+    """Test that the JSON representation refers to correct class.
+
+    Make sure that the JSON representation of a CachingFileSystem refers to the
+    CachingFileSystem, not to the underlying filesystem.
+    """
+    import json
+
+    from fsspec.implementations.local import LocalFileSystem
+
+    lfs = LocalFileSystem()
+    cfs = CachingFileSystem(fs=lfs, cache_storage="raspberry")
+    D = json.loads(cfs.to_json())
+    assert D["cls"].endswith("CachingFileSystem")
