@@ -1,4 +1,5 @@
 import pytest
+
 import fsspec
 
 pytest.importorskip("distributed")
@@ -24,6 +25,6 @@ def cli(tmpdir):
 
 def test_basic(cli):
 
-    fs = fsspec.filesystem("dask", remote_protocol="memory")
-    assert fs.ls("") == ["afile"]
-    assert fs.cat("afile") == b"data"
+    fs = fsspec.filesystem("dask", target_protocol="memory")
+    assert fs.ls("") == ["/afile"]
+    assert fs.cat("/afile") == b"data"
