@@ -1,15 +1,18 @@
 import io
 import logging
 import os
+import threading
 import warnings
-from distutils.version import LooseVersion
+import weakref
 from errno import ESPIPE
-from hashlib import sha256
 from glob import has_magic
+from hashlib import sha256
 
+from .callbacks import _DEFAULT_CALLBACK
+from .config import apply_config, conf
 from .dircache import DirCache
 from .transaction import Transaction
-from .utils import read_block, tokenize, stringify_path, other_paths
+from .utils import _unstrip_protocol, other_paths, read_block, stringify_path, tokenize
 
 logger = logging.getLogger("fsspec")
 
@@ -40,14 +43,21 @@ class _Cached(type):
         # Note: we intentionally create a reference here, to avoid garbage
         # collecting instances when all other references are gone. To really
         # delete a FileSystem, the cache must be cleared.
-        cls._cache = {}
+        if conf.get("weakref_instance_cache"):  # pragma: no cover
+            # debug option for analysing fork/spawn conditions
+            cls._cache = weakref.WeakValueDictionary()
+        else:
+            cls._cache = {}
         cls._pid = os.getpid()
 
     def __call__(cls, *args, **kwargs):
+        kwargs = apply_config(cls, kwargs)
         extra_tokens = tuple(
             getattr(cls, attr, None) for attr in cls._extra_tokenize_attributes
         )
-        token = tokenize(cls, *args, *extra_tokens, **kwargs)
+        token = tokenize(
+            cls, cls._pid, threading.get_ident(), *args, *extra_tokens, **kwargs
+        )
         skip = kwargs.pop("skip_instance_cache", False)
         if os.getpid() != cls._pid:
             cls._cache.clear()
@@ -70,19 +80,7 @@ class _Cached(type):
             return obj
 
 
-try:  # optionally derive from pyarrow's FileSystem, if available
-    import pyarrow as pa
-except ImportError:
-    up = object
-else:
-    # only derive from the legacy pyarrow's FileSystem for older pyarrow versions
-    if LooseVersion(pa.__version__) < LooseVersion("2.0"):
-        up = pa.filesystem.DaskFileSystem
-    else:
-        up = object
-
-
-class AbstractFileSystem(up, metaclass=_Cached):
+class AbstractFileSystem(metaclass=_Cached):
     """
     An abstract super-class for pythonic file-systems
 
@@ -240,7 +238,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         # Not necessary to implement invalidation mechanism, may have no cache.
         # But if have, you should call this method of parent class from your
         # subclass to ensure expiring caches after transacations correctly.
-        # See the implementaion of FTPFileSystem in ftp.py
+        # See the implementation of FTPFileSystem in ftp.py
         if self._intrans:
             self._invalidated_caches_in_transaction.append(path)
 
@@ -274,7 +272,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         path: str
             leaf directory name
         exist_ok: bool (False)
-            If True, will error if the target already exists
+            If False, will error if the target already exists
         """
         pass  # not necessary to implement, may not have directories
 
@@ -330,12 +328,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
         but contains nothing), None if not in cache.
         """
         parent = self._parent(path)
+        if path.rstrip("/") in self.dircache:
+            return self.dircache[path.rstrip("/")]
         try:
-            return self.dircache[path]
-        except KeyError:
-            pass
-        try:
-            files = [f for f in self.dircache[parent] if f["name"] == path]
+            files = [
+                f
+                for f in self.dircache[parent]
+                if f["name"] == path
+                or (f["name"] == path.rstrip("/") and f["type"] == "directory")
+            ]
             if len(files) == 0:
                 # parent dir was listed but did not contain this file
                 raise FileNotFoundError(path)
@@ -370,7 +371,9 @@ class AbstractFileSystem(up, metaclass=_Cached):
         try:
             listing = self.ls(path, detail=True, **kwargs)
         except (FileNotFoundError, IOError):
-            return [], [], []
+            if detail:
+                return path, {}, {}
+            return path, [], []
 
         for info in listing:
             # each info name must be at least [path]/part , but here
@@ -419,11 +422,11 @@ class AbstractFileSystem(up, metaclass=_Cached):
         path = self._strip_protocol(path)
         out = dict()
         detail = kwargs.pop("detail", False)
-        for path, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
+        for _, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
             if withdirs:
                 files.update(dirs)
             out.update({info["name"]: info for name, info in files.items()})
-        if self.isfile(path) and path not in out:
+        if not out and self.isfile(path):
             # walk works on directories, but find should also return [path]
             # when path happens to be a file
             out[path] = {}
@@ -508,7 +511,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
             depth = None if "**" in path else path[ind2 + 1 :].count("/") + 1
         else:
             root = ""
-            depth = None if "**" in path else 1
+            depth = None if "**" in path else path[ind + 1 :].count("/") + 1
 
         allpaths = self.find(root, maxdepth=depth, withdirs=True, detail=True, **kwargs)
         # Escape characters special to python regex, leaving our supported
@@ -547,15 +550,20 @@ class AbstractFileSystem(up, metaclass=_Cached):
         else:
             return list(out)
 
-    def exists(self, path):
+    def exists(self, path, **kwargs):
         """Is there a file at the given path"""
         try:
-            self.info(path)
+            self.info(path, **kwargs)
             return True
         except:  # noqa: E722
             # any exception allowed bar FileNotFoundError?
             return False
 
+    def lexists(self, path, **kwargs):
+        """If there is a file at the given path (including
+        broken links)"""
+        return self.exists(path)
+
     def info(self, path, **kwargs):
         """Give details of entry at path
 
@@ -607,6 +615,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Size in bytes of file"""
         return self.info(path).get("size", None)
 
+    def sizes(self, paths):
+        """Size in bytes of each file in a list of paths"""
+        return [self.size(p) for p in paths]
+
     def isdir(self, path):
         """Is this entry directory-like?"""
         try:
@@ -621,9 +633,30 @@ class AbstractFileSystem(up, metaclass=_Cached):
         except:  # noqa: E722
             return False
 
-    def cat_file(self, path):
-        """ Get the content of a file """
-        return self.open(path, "rb").read()
+    def cat_file(self, path, start=None, end=None, **kwargs):
+        """Get the content of a file
+
+        Parameters
+        ----------
+        path: URL of file on this filesystems
+        start, end: int
+            Bytes limits of the read. If negative, backwards from end,
+            like usual python slices. Either can be None for start or
+            end of file, respectively
+        kwargs: passed to ``open()``.
+        """
+        # explicitly set buffering off?
+        with self.open(path, "rb", **kwargs) as f:
+            if start is not None:
+                if start >= 0:
+                    f.seek(start)
+                else:
+                    f.seek(max(0, f.size + start))
+            if end is not None:
+                if end < 0:
+                    end = f.size + end
+                return f.read(end - f.tell())
+            return f.read()
 
     def pipe_file(self, path, value, **kwargs):
         """Set the bytes of given file"""
@@ -634,6 +667,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Put value into path
 
         (counterpart to ``cat``)
+
         Parameters
         ----------
         path: string or dict(str, bytes)
@@ -651,18 +685,39 @@ class AbstractFileSystem(up, metaclass=_Cached):
         else:
             raise ValueError("path must be str or dict")
 
+    def cat_ranges(self, paths, starts, ends, max_gap=None, **kwargs):
+        if max_gap is not None:
+            raise NotImplementedError
+        if not isinstance(paths, list):
+            raise TypeError
+        if not isinstance(starts, list):
+            starts = [starts] * len(paths)
+        if not isinstance(ends, list):
+            ends = [starts] * len(paths)
+        if len(starts) != len(paths) or len(ends) != len(paths):
+            raise ValueError
+        return [self.cat_file(p, s, e) for p, s, e in zip(paths, starts, ends)]
+
     def cat(self, path, recursive=False, on_error="raise", **kwargs):
         """Fetch (potentially multiple) paths' contents
 
-        Returns a dict of {path: contents} if there are multiple paths
-        or the path has been otherwise expanded
-
+        Parameters
+        ----------
+        recursive: bool
+            If True, assume the path(s) are directories, and get all the
+            contained files
         on_error : "raise", "omit", "return"
             If raise, an underlying exception will be raised (converted to KeyError
             if the type is in self.missing_exceptions); if omit, keys with exception
             will simply not be included in the output; if "return", all keys are
             included in the output, but the value will be bytes or an exception
             instance.
+        kwargs: passed to cat_file
+
+        Returns
+        -------
+        dict of {path: contents} if there are multiple paths
+        or the path has been otherwise expanded
         """
         paths = self.expand_path(path, recursive=recursive)
         if (
@@ -681,21 +736,24 @@ class AbstractFileSystem(up, metaclass=_Cached):
                         out[path] = e
             return out
         else:
-            return self.cat_file(paths[0])
+            return self.cat_file(paths[0], **kwargs)
 
-    def get_file(self, rpath, lpath, **kwargs):
+    def get_file(self, rpath, lpath, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy single remote file to local"""
         if self.isdir(rpath):
             os.makedirs(lpath, exist_ok=True)
-        else:
-            with self.open(rpath, "rb", **kwargs) as f1:
-                with open(lpath, "wb") as f2:
-                    data = True
-                    while data:
-                        data = f1.read(self.blocksize)
-                        f2.write(data)
+            return None
 
-    def get(self, rpath, lpath, recursive=False, **kwargs):
+        with self.open(rpath, "rb", **kwargs) as f1:
+            callback.set_size(getattr(f1, "size", None))
+            with open(lpath, "wb") as f2:
+                data = True
+                while data:
+                    data = f1.read(self.blocksize)
+                    segment_len = f2.write(data)
+                    callback.relative_update(segment_len)
+
+    def get(self, rpath, lpath, recursive=False, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy file(s) to local.
 
         Copies a specific file or tree of files (if recursive=True). If lpath
@@ -711,23 +769,31 @@ class AbstractFileSystem(up, metaclass=_Cached):
             lpath = make_path_posix(lpath)
         rpaths = self.expand_path(rpath, recursive=recursive)
         lpaths = other_paths(rpaths, lpath)
-        for lpath, rpath in zip(lpaths, rpaths):
+
+        callback.set_size(len(lpaths))
+        for lpath, rpath in callback.wrap(zip(lpaths, rpaths)):
+            callback.branch(rpath, lpath, kwargs)
             self.get_file(rpath, lpath, **kwargs)
 
-    def put_file(self, lpath, rpath, **kwargs):
+    def put_file(self, lpath, rpath, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy single file to remote"""
         if os.path.isdir(lpath):
             self.makedirs(rpath, exist_ok=True)
-        else:
-            with open(lpath, "rb") as f1:
-                self.mkdirs(os.path.dirname(rpath), exist_ok=True)
-                with self.open(rpath, "wb", **kwargs) as f2:
-                    data = True
-                    while data:
-                        data = f1.read(self.blocksize)
-                        f2.write(data)
+            return None
 
-    def put(self, lpath, rpath, recursive=False, **kwargs):
+        with open(lpath, "rb") as f1:
+            size = f1.seek(0, 2)
+            callback.set_size(size)
+            f1.seek(0)
+
+            self.mkdirs(self._parent(os.fspath(rpath)), exist_ok=True)
+            with self.open(rpath, "wb", **kwargs) as f2:
+                while f1.tell() < size:
+                    data = f1.read(self.blocksize)
+                    segment_len = f2.write(data)
+                    callback.relative_update(segment_len)
+
+    def put(self, lpath, rpath, recursive=False, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy file(s) from local.
 
         Copies a specific file or tree of files (if recursive=True). If rpath
@@ -736,7 +802,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
         Calls put_file for each source.
         """
-        from .implementations.local import make_path_posix, LocalFileSystem
+        from .implementations.local import LocalFileSystem, make_path_posix
 
         rpath = (
             self._strip_protocol(rpath)
@@ -747,18 +813,22 @@ class AbstractFileSystem(up, metaclass=_Cached):
             lpath = make_path_posix(lpath)
         fs = LocalFileSystem()
         lpaths = fs.expand_path(lpath, recursive=recursive)
-        rpaths = other_paths(lpaths, rpath)
+        rpaths = other_paths(
+            lpaths, rpath, exists=isinstance(rpath, str) and self.isdir(rpath)
+        )
 
-        for lpath, rpath in zip(lpaths, rpaths):
+        callback.set_size(len(rpaths))
+        for lpath, rpath in callback.wrap(zip(lpaths, rpaths)):
+            callback.branch(lpath, rpath, kwargs)
             self.put_file(lpath, rpath, **kwargs)
 
     def head(self, path, size=1024):
-        """ Get the first ``size`` bytes from file """
+        """Get the first ``size`` bytes from file"""
         with self.open(path, "rb") as f:
             return f.read(size)
 
     def tail(self, path, size=1024):
-        """ Get the last ``size`` bytes from file """
+        """Get the last ``size`` bytes from file"""
         with self.open(path, "rb") as f:
             f.seek(max(-size, -f.size), 2)
             return f.read()
@@ -766,18 +836,36 @@ class AbstractFileSystem(up, metaclass=_Cached):
     def cp_file(self, path1, path2, **kwargs):
         raise NotImplementedError
 
-    def copy(self, path1, path2, recursive=False, **kwargs):
-        """ Copy within two locations in the filesystem"""
+    def copy(self, path1, path2, recursive=False, on_error=None, **kwargs):
+        """Copy within two locations in the filesystem
+
+        on_error : "raise", "ignore"
+            If raise, any not-found exceptions will be raised; if ignore any
+            not-found exceptions will cause the path to be skipped; defaults to
+            raise unless recursive is true, where the default is ignore
+        """
+        if on_error is None and recursive:
+            on_error = "ignore"
+        elif on_error is None:
+            on_error = "raise"
+
         paths = self.expand_path(path1, recursive=recursive)
         path2 = other_paths(paths, path2)
         for p1, p2 in zip(paths, path2):
-            self.cp_file(p1, p2, **kwargs)
+            try:
+                self.cp_file(p1, p2, **kwargs)
+            except FileNotFoundError:
+                if on_error == "raise":
+                    raise
 
     def expand_path(self, path, recursive=False, maxdepth=None):
-        """Turn one or more globs or directories into a list of all matching files"""
+        """Turn one or more globs or directories into a list of all matching paths
+        to files or directories."""
         if isinstance(path, str):
             out = self.expand_path([path], recursive, maxdepth)
         else:
+            # reduce depth on each recursion level unless None or 0
+            maxdepth = maxdepth if not maxdepth else maxdepth - 1
             out = set()
             path = [self._strip_protocol(p) for p in path]
             for p in path:
@@ -785,10 +873,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
                     bit = set(self.glob(p))
                     out |= bit
                     if recursive:
-                        out += self.expand_path(p)
+                        out |= set(
+                            self.expand_path(
+                                list(bit), recursive=recursive, maxdepth=maxdepth
+                            )
+                        )
                     continue
                 elif recursive:
-                    rec = set(self.find(p, withdirs=True))
+                    rec = set(self.find(p, maxdepth=maxdepth, withdirs=True))
                     out |= rec
                 if p not in out and (recursive is False or self.exists(p)):
                     # should only check once, for the root
@@ -798,7 +890,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         return list(sorted(out))
 
     def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
-        """ Move file(s) from one location to another """
+        """Move file(s) from one location to another"""
         self.copy(path1, path2, recursive=recursive, maxdepth=maxdepth)
         self.rm(path1, recursive=recursive)
 
@@ -846,7 +938,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         block_size=None,
         autocommit=True,
         cache_options=None,
-        **kwargs
+        **kwargs,
     ):
         """Return raw bytes-mode file-like from the file-system"""
         return AbstractBufferedFile(
@@ -856,10 +948,18 @@ class AbstractFileSystem(up, metaclass=_Cached):
             block_size,
             autocommit,
             cache_options=cache_options,
-            **kwargs
+            **kwargs,
         )
 
-    def open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
+    def open(
+        self,
+        path,
+        mode="rb",
+        block_size=None,
+        cache_options=None,
+        compression=None,
+        **kwargs,
+    ):
         """
         Return a file-like object from the filesystem
 
@@ -876,6 +976,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
             Some indication of buffering - this is a value in bytes
         cache_options : dict, optional
             Extra arguments to pass through to the cache.
+        compression: string or None
+            If given, open file using compression codec. Can either be a compression
+            name (a key in ``fsspec.compression.compr``) or "infer" to guess the
+            compression from the filename suffix.
         encoding, errors, newline: passed on to TextIOWrapper for text mode
         """
         import io
@@ -890,7 +994,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 if k in kwargs
             }
             return io.TextIOWrapper(
-                self.open(path, mode, block_size, **kwargs), **text_kwargs
+                self.open(
+                    path,
+                    mode,
+                    block_size=block_size,
+                    cache_options=cache_options,
+                    compression=compression,
+                    **kwargs,
+                ),
+                **text_kwargs,
             )
         else:
             ac = kwargs.pop("autocommit", not self._intrans)
@@ -900,9 +1012,17 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 block_size=block_size,
                 autocommit=ac,
                 cache_options=cache_options,
-                **kwargs
+                **kwargs,
             )
-            if not ac:
+            if compression is not None:
+                from fsspec.compression import compr
+                from fsspec.core import get_compression
+
+                compression = get_compression(path, compression)
+                compress = compr[compression]
+                f = compress(f, mode=mode[0])
+
+            if not ac and "r" not in mode:
                 self.transaction.files.append(f)
             return f
 
@@ -995,7 +1115,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         return json.dumps(
             dict(
                 **{"cls": cls, "protocol": proto, "args": self.storage_args},
-                **self.storage_options
+                **self.storage_options,
             )
         )
 
@@ -1014,9 +1134,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         -------
         file system instance, not necessarily of this particular class.
         """
-        from .registry import _import_class, get_filesystem_class
         import json
 
+        from .registry import _import_class, get_filesystem_class
+
         dic = json.loads(blob)
         protocol = dic.pop("protocol")
         try:
@@ -1035,7 +1156,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
     def get_mapper(self, root, check=False, create=False):
         """Create key/value store based on this file-system
 
-        Makes a MutibleMapping interface to the FS at the given root path.
+        Makes a MutableMapping interface to the FS at the given root path.
         See ``fsspec.mapping.FSMap`` for further details.
         """
         from .mapping import FSMap
@@ -1069,47 +1190,47 @@ class AbstractFileSystem(up, metaclass=_Cached):
     # Aliases
 
     def makedir(self, path, create_parents=True, **kwargs):
-        """Alias of :ref:`FilesystemSpec.mkdir`."""
+        """Alias of `AbstractFileSystem.mkdir`."""
         return self.mkdir(path, create_parents=create_parents, **kwargs)
 
     def mkdirs(self, path, exist_ok=False):
-        """Alias of :ref:`FilesystemSpec.makedirs`."""
+        """Alias of `AbstractFileSystem.makedirs`."""
         return self.makedirs(path, exist_ok=exist_ok)
 
     def listdir(self, path, detail=True, **kwargs):
-        """Alias of :ref:`FilesystemSpec.ls`."""
+        """Alias of `AbstractFileSystem.ls`."""
         return self.ls(path, detail=detail, **kwargs)
 
     def cp(self, path1, path2, **kwargs):
-        """Alias of :ref:`FilesystemSpec.copy`."""
+        """Alias of `AbstractFileSystem.copy`."""
         return self.copy(path1, path2, **kwargs)
 
     def move(self, path1, path2, **kwargs):
-        """Alias of :ref:`FilesystemSpec.mv`."""
+        """Alias of `AbstractFileSystem.mv`."""
         return self.mv(path1, path2, **kwargs)
 
     def stat(self, path, **kwargs):
-        """Alias of :ref:`FilesystemSpec.info`."""
+        """Alias of `AbstractFileSystem.info`."""
         return self.info(path, **kwargs)
 
     def disk_usage(self, path, total=True, maxdepth=None, **kwargs):
-        """Alias of :ref:`FilesystemSpec.du`."""
+        """Alias of `AbstractFileSystem.du`."""
         return self.du(path, total=total, maxdepth=maxdepth, **kwargs)
 
     def rename(self, path1, path2, **kwargs):
-        """Alias of :ref:`FilesystemSpec.mv`."""
+        """Alias of `AbstractFileSystem.mv`."""
         return self.mv(path1, path2, **kwargs)
 
     def delete(self, path, recursive=False, maxdepth=None):
-        """Alias of :ref:`FilesystemSpec.rm`."""
+        """Alias of `AbstractFileSystem.rm`."""
         return self.rm(path, recursive=recursive, maxdepth=maxdepth)
 
     def upload(self, lpath, rpath, recursive=False, **kwargs):
-        """Alias of :ref:`FilesystemSpec.put`."""
+        """Alias of `AbstractFileSystem.put`."""
         return self.put(lpath, rpath, recursive=recursive, **kwargs)
 
     def download(self, rpath, lpath, recursive=False, **kwargs):
-        """Alias of :ref:`FilesystemSpec.get`."""
+        """Alias of `AbstractFileSystem.get`."""
         return self.get(rpath, lpath, recursive=recursive, **kwargs)
 
     def sign(self, path, expiration=100, **kwargs):
@@ -1132,14 +1253,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
         Raises
         ------
-        NotImplementedError : if method is not implemented for a fileystem
+        NotImplementedError : if method is not implemented for a filesystem
         """
         raise NotImplementedError("Sign is not implemented for this filesystem")
 
     def _isfilestore(self):
         # Originally inherited from pyarrow DaskFileSystem. Keeping this
         # here for backwards compatibility as long as pyarrow uses its
-        # legacy ffspec-compatible filesystems and thus accepts fsspec
+        # legacy fsspec-compatible filesystems and thus accepts fsspec
         # filesystems as well
         return False
 
@@ -1150,10 +1271,11 @@ class AbstractBufferedFile(io.IOBase):
     In the case that the backend does not provide a pythonic file-like object
     already, this class contains much of the logic to build one. The only
     methods that need to be overridden are ``_upload_chunk``,
-    ``_initate_upload`` and ``_fetch_range``.
+    ``_initiate_upload`` and ``_fetch_range``.
     """
 
     DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
+    _details = None
 
     def __init__(
         self,
@@ -1164,7 +1286,8 @@ class AbstractBufferedFile(io.IOBase):
         autocommit=True,
         cache_type="readahead",
         cache_options=None,
-        **kwargs
+        size=None,
+        **kwargs,
     ):
         """
         Template for files with buffered reading and writing
@@ -1187,6 +1310,8 @@ class AbstractBufferedFile(io.IOBase):
         cache_options : dict
             Additional options passed to the constructor for the cache specified
             by `cache_type`.
+        size: int
+            If given and in read mode, suppressed having to look up the file size
         kwargs:
             Gets stored as self.kwargs
         """
@@ -1220,9 +1345,10 @@ class AbstractBufferedFile(io.IOBase):
         if mode not in {"ab", "rb", "wb"}:
             raise NotImplementedError("File mode not supported")
         if mode == "rb":
-            if not hasattr(self, "details"):
-                self.details = fs.info(path)
-            self.size = self.details["size"]
+            if size is not None:
+                self.size = size
+            else:
+                self.size = self.details["size"]
             self.cache = caches[cache_type](
                 self.blocksize, self._fetch_range, self.size, **cache_options
             )
@@ -1232,6 +1358,21 @@ class AbstractBufferedFile(io.IOBase):
             self.forced = False
             self.location = None
 
+    @property
+    def details(self):
+        if self._details is None:
+            self._details = self.fs.info(self.path)
+        return self._details
+
+    @details.setter
+    def details(self, value):
+        self._details = value
+        self.size = value["size"]
+
+    @property
+    def full_name(self):
+        return _unstrip_protocol(self.path, self.fs)
+
     @property
     def closed(self):
         # get around this attr being read-only in IOBase
@@ -1259,14 +1400,14 @@ class AbstractBufferedFile(io.IOBase):
         """Throw away temporary file"""
 
     def info(self):
-        """ File information about this path """
+        """File information about this path"""
         if "r" in self.mode:
             return self.details
         else:
             raise ValueError("Info not available while writing")
 
     def tell(self):
-        """ Current file location """
+        """Current file location"""
         return self.loc
 
     def seek(self, loc, whence=0):
@@ -1351,7 +1492,11 @@ class AbstractBufferedFile(io.IOBase):
         if self.offset is None:
             # Initialize a multipart upload
             self.offset = 0
-            self._initiate_upload()
+            try:
+                self._initiate_upload()
+            except:  # noqa: E722
+                self.closed = True
+                raise
 
         if self._upload_chunk(final=force) is not False:
             self.offset += self.buffer.seek(0, 2)
@@ -1369,7 +1514,7 @@ class AbstractBufferedFile(io.IOBase):
         # may not yet have been initialized, may need to call _initialize_upload
 
     def _initiate_upload(self):
-        """ Create remote file/upload """
+        """Create remote file/upload"""
         pass
 
     def _fetch_range(self, start, end):
@@ -1474,6 +1619,8 @@ class AbstractBufferedFile(io.IOBase):
 
         Finalizes writes, discards cache
         """
+        if getattr(self, "_unclosable", False):
+            return
         if self.closed:
             return
         if self.mode == "rb":
