@@ -9,7 +9,7 @@ import tempfile
 import pytest
 
 import fsspec
-from fsspec.implementations.memory import MemoryFileSystem, MemoryFile
+from fsspec.implementations.memory import MemoryFile, MemoryFileSystem
 
 
 def test_idempotent():
@@ -33,9 +33,8 @@ def test_pickle():
 
 
 def test_class_methods():
-    assert MemoryFileSystem._strip_protocol("memory::stuff") == "stuff"
-    assert MemoryFileSystem._strip_protocol("memory://stuff") == "stuff"
-    assert MemoryFileSystem._strip_protocol("stuff") == "stuff"
+    assert MemoryFileSystem._strip_protocol("memory://stuff") == "/stuff"
+    assert MemoryFileSystem._strip_protocol("stuff") == "/stuff"
     assert MemoryFileSystem._strip_protocol("other://stuff") == "other://stuff"
 
     assert MemoryFileSystem._get_kwargs_from_urls("memory://user@thing") == {}
@@ -146,7 +145,7 @@ def test_pipe_cat():
     fs.pipe("afile", b"contents")
     assert fs.cat("afile") == b"contents"
 
-    data = {"bfile": b"more", "cfile": b"stuff"}
+    data = {"/bfile": b"more", "/cfile": b"stuff"}
     fs.pipe(data)
     assert fs.cat(list(data)) == data
 
@@ -190,7 +189,7 @@ def test_chained_fs():
 
 @pytest.mark.xfail(reason="see issue #334", strict=True)
 def test_multilevel_chained_fs():
-    """This test reproduces intake/filesystem_spec#334"""
+    """This test reproduces fsspec/filesystem_spec#334"""
     import zipfile
 
     d1 = tempfile.mkdtemp()
@@ -216,7 +215,7 @@ def test_multilevel_chained_fs():
 
 @pytest.mark.skipif(sys.version_info < (3, 7), reason="no seek in old zipfile")
 def test_multilevel_chained_fs_zip_zip_file():
-    """This test reproduces intake/filesystem_spec#334"""
+    """This test reproduces fsspec/filesystem_spec#334"""
     import zipfile
 
     d1 = tempfile.mkdtemp()
@@ -324,3 +323,11 @@ def test_chained_fo():
     with of[0] as f:
         assert f.read() == b"test"
     assert "afile" in os.listdir(d3)
+
+
+def test_url_to_fs():
+    url = "memory://a.txt"
+    fs, url2 = fsspec.core.url_to_fs(url)
+
+    assert isinstance(fs, MemoryFileSystem)
+    assert url2 == "/a.txt"
