@@ -1,7 +1,9 @@
 """Tests abstract buffered file API, using FTP implementation"""
 import pickle
 import sys
+
 import pytest
+
 from fsspec.implementations.tests.test_ftp import FTPFileSystem
 
 data = b"hello" * 10000
@@ -9,7 +11,7 @@ data = b"hello" * 10000
 
 @pytest.mark.xfail(
     sys.version_info < (3, 6),
-    reason="py35 error, see https://github.com/intake/filesystem_spec/issues/147",
+    reason="py35 error, see https://github.com/fsspec/filesystem_spec/issues/147",
 )
 def test_pickle(ftp_writable):
     host, port, user, pw = ftp_writable
@@ -164,7 +166,7 @@ def test_read_block(ftp_writable):
 def test_with_gzip(ftp_writable):
     import gzip
 
-    data = b"some compressable stuff"
+    data = b"some compressible stuff"
     host, port, user, pw = ftp_writable
     fs = FTPFileSystem(host=host, port=port, username=user, password=pw)
     fn = "/myfile"
@@ -177,6 +179,14 @@ def test_with_gzip(ftp_writable):
         assert gf.read() == data
 
 
+def test_auto_compression(m):
+    fs = m
+    with fs.open("myfile.gz", mode="wt", compression="infer") as f:
+        f.write("text")
+    with fs.open("myfile.gz", mode="rt", compression="infer") as f:
+        assert f.read() == "text"
+
+
 def test_with_zip(ftp_writable):
     import zipfile
 
